class HasDisable:
    def __init__(self) -> None:
        self.disable = False
        """Set to true to make the job fail if it is executed. Keeping disabled jobs instead of deleting them is useful
        to keep data lineage of your legacy data while making sure the job won't be accidentally run. Disabled jobs will
        be colored in grey in the dependency graph generated by the `show_graph` command."""
